// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
#include <ntddk.h>
#include "datapipe.h"
#include "memtrack.h"

struct datapipe_entry {
	struct datapipe_entry *next;
	ULONG size;
	char data[];
};

void datapipe_init(datapipe_t *dp)
{
	dp->first = dp->last = NULL;
	KeInitializeSpinLock(&dp->guard);
}

void datapipe_free(datapipe_t *dp)
{
	struct datapipe_entry *de;
	KIRQL irql;
	KeAcquireSpinLock(&dp->guard, &irql);

	for (de = dp->first; de;) {
		struct datapipe_entry *de2 = de->next;
		free(de);
		de = de2;
	}

	KeReleaseSpinLock(&dp->guard, irql);
	memset(dp, 0, sizeof(*dp));
}

ULONG datapipe_peek(datapipe_t *dp)
{
	ULONG result;
	KIRQL irql;
	KeAcquireSpinLock(&dp->guard, &irql);
	
	if (!dp->first) result = 0; else result = dp->first->size;
	
	KeReleaseSpinLock(&dp->guard, irql);
	return result;
}

ULONG datapipe_get(datapipe_t *dp, char *buf, ULONG size)
{
	struct datapipe_entry *de;
	ULONG result = 0;
	KIRQL irql;

	KeAcquireSpinLock(&dp->guard, &irql);
	de = dp->first;

	if (!de) goto done;
	if (de->size > size) goto done;

	dp->first = de->next;

	memcpy(buf, de->data, de->size);
	result = de->size;

	free(de);
done:
	KeReleaseSpinLock(&dp->guard, irql);
	return result;
}

NTSTATUS datapipe_push(datapipe_t *dp, const char *data, ULONG size)
{
	struct datapipe_entry *de;
	KIRQL irql;

	de = (struct datapipe_entry *)malloc(sizeof(*de) + size);
	if (!de) return STATUS_NO_MEMORY;

	de->next = NULL;
	de->size = size;
	memcpy(de->data, data, size);

	KeAcquireSpinLock(&dp->guard, &irql);

	if (dp->last) dp->last->next = de; else dp->last = de;
	if (!dp->first) dp->first = dp->last;
	
	KeReleaseSpinLock(&dp->guard, irql);

	return STATUS_SUCCESS;
}
